/**
  *  \file client/dialogs/simplanetedit.cc
  *  \brief Simulator Planet Editor
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/simplanetedit.h"
#include "client/dialogs/helpdlg.h"
#include "client/dialogs/simedit.h"
#include "client/dialogs/simplanettorps.h"
#include "client/dlg-fcode.h"
#include "client/wtile.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/formula.h"
#include "game/racename.h"
#include "game/sim-game.h"
#include "game/sim.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/dialogs/simfunction.h"
#include "ui/widgets/quit.h"

namespace {
    class WSimPlanetEditor;

    class WSimPlanetPlanetTile : public WTile {
        // name N, owner O, id I, fcode F, defense posts PD, level L, Abilities A
        GSimPlanet& planet;
        ObjectHolder h;

        void init();
        void drawContent(UITableWriter& writer, bool with_content);
     public:
        WSimPlanetPlanetTile(GSimPlanet& planet);

        void drawData(GfxCanvas& can);
    };

    class WSimPlanetBaseTile : public WTile {
        // beam tech B, fighters G, sb defense S, torps T, weapons W
        GSimPlanet& planet;
        ObjectHolder h;

        void init();
        void drawContent(UITableWriter& writer, bool with_content);
     public:
        WSimPlanetBaseTile(GSimPlanet& planet);

        void drawData(GfxCanvas& can);
    };

    class WSimPlanetEditButton : public UIButton {
        GSimPlanet& planet;
     public:
        WSimPlanetEditButton(GSimPlanet& planet, const char* title, UIKey key);
        void update();
    };

    class WSimPlanetEditor : public UIWindow {
        GSimPlanet& planet;
        ObjectHolder holder;
        Signal0 sig_edit;
        GSimulatorGameInterface& game_if;

        void init();
     public:
        WSimPlanetEditor(GSimulatorGameInterface& game_if, GSimPlanet& planet);

        void editName();
        void editOwner();
        void editId();
        void editFCode();
        void editDefense();
        void editPopulation();
        void editBaseBeamTech();
        void editBaseFighters();
        void editBaseDefense();
        void editBaseTorpTech();
        void editBaseTorps();

        bool handleEvent(const UIEvent& e, bool second_pass);
        void update();
    };
}

static void
setPlanetName(GSimulatorGameInterface& game_if, GSimPlanet& pl)
{
    pl.setName(game_if.getPlanetName(pl.getId()));
}

static inline bool
isExpGame()
{
    return config.NumExperienceLevels() > 0;
}

/************************** WSimPlanetPlanetTile *************************/

WSimPlanetPlanetTile::WSimPlanetPlanetTile(GSimPlanet& planet)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(30, isExpGame() ? 7 : 6)),
      planet(planet)
{
    init();
}

void
WSimPlanetPlanetTile::init()
{
    setHeading(_("Planet:"));
    setDisplayState(sNormal);

    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(em,       0, em1, em1), h.add(new WSimPlanetEditButton(planet, "N", 'n')));
    addControl(GfxRect(em,      em, em1, em1), h.add(new WSimPlanetEditButton(planet, "O", 'o')));
    addControl(GfxRect(em,  2 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "I", 'i')));
    addControl(GfxRect(em,  3 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "F", 'f')));
    addControl(GfxRect(em,  4 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "D", 'd')));
    addControl(GfxRect(0,   4 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "P", 'p')));
    addControl(GfxRect(em,  5 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "A", 'a')));
    if (isExpGame())
        addControl(GfxRect(em, 6 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "L", 'l')));
}

void
WSimPlanetPlanetTile::drawContent(UITableWriter& writer, bool with_content)
{
    writer.newLColumn();
    writer.writeFixedCell(UIColor::tc_Static, _("Name:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Owner:"));
    writer.writeFixedCell(UIColor::tc_Static, _("ID:"));
    writer.writeFixedCell(UIColor::tc_Static, _("FCode:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Defense Posts:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Abilities:"));
    if (isExpGame())
        writer.writeFixedCell(UIColor::tc_Static, _("Level:"));

    if (with_content) {
        writer.newLColumn();
        writer.writeVariableCell(UIColor::tc_Green, planet.getName());
        writer.writeVariableCell(UIColor::tc_Green, player_racenames.getShortName(planet.getOwner()));
        writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getId()));
        writer.writeVariableCell(UIColor::tc_Green, planet.getFCode());
        writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getDefense()));
        writer.writeVariableCell(planet.hasAnyNonstandardFunction() ? UIColor::tc_Yellow : UIColor::tc_Green,
                                 describeFunctions(planet));
        if (isExpGame()) {
            writer.writeVariableCell(UIColor::tc_Green, getExperienceLevelName(planet.getExperienceLevel()));
        }
    }
    writer.finish();
}

void
WSimPlanetPlanetTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter writer(ctx, getInnerSize(10, font_heights[FONT_NORMAL]*2), FONT_NORMAL);
    drawContent(writer, true);
}

/*************************** WSimPlanetBaseTile **************************/

WSimPlanetBaseTile::WSimPlanetBaseTile(GSimPlanet& planet)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(30, 5)),
      planet(planet)
{
    init();
}

void
WSimPlanetBaseTile::init()
{
    setHeading(_("Starbase:"));
    setDisplayState(sNormal);

    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(0,    0, em1, em1), h.add(new WSimPlanetEditButton(planet, "B", 'b')));
    addControl(GfxRect(0,   em, em1, em1), h.add(new WSimPlanetEditButton(planet, "G", 'g')));
    addControl(GfxRect(0, 2*em, em1, em1), h.add(new WSimPlanetEditButton(planet, "S", 's')));
    addControl(GfxRect(0, 3*em, em1, em1), h.add(new WSimPlanetEditButton(planet, "T", 't')));
    addControl(GfxRect(0, 4*em, em1, em1), h.add(new WSimPlanetEditButton(planet, "W", 'w')));
}

void
WSimPlanetBaseTile::drawContent(UITableWriter& writer, bool with_content)
{
    writer.newLColumn();
    writer.writeFixedCell(UIColor::tc_Static, _("Beam Tech:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Fighters:"));
    writer.writeFixedCell(UIColor::tc_Static, _("SB Defense:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Torp Tech:"));
    writer.writeFixedCell(UIColor::tc_Static, _("# Torpedoes:"));

    if (with_content) {
        writer.newLColumn();
        if (planet.getBaseBeamTech() != 0) {
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseBeamTech()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseFighters()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseDefense()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseTorpTech()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseTorpsAsType(planet.getBaseTorpTech())));
        } else {
            writer.writeVariableCell(UIColor::tc_Green, _("(no starbase)"));
            for (int i = 0; i < 4; ++i)
                writer.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    writer.finish();
}

void
WSimPlanetBaseTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter writer(ctx, getInnerSize(10, font_heights[FONT_NORMAL]), FONT_NORMAL);
    drawContent(writer, true);
}

/************************** WSimPlanetEditButton *************************/

WSimPlanetEditButton::WSimPlanetEditButton(GSimPlanet& planet, const char* title, UIKey key)
    : UIButton(title, key), planet(planet)
{
    setState(bf_SmallSize, true);
    update();
}

void
WSimPlanetEditButton::update()
{
    bool enabled;
    switch (getKey()) {
     case 'g':
     case 's':
     case 't':
     case 'w':
        enabled = (planet.getBaseBeamTech() != 0);
        break;
     default:
        enabled = true;
        break;
    }
    setState(st_Disabled, !enabled);
}

/**************************** WSimPlanetEditor ***************************/

WSimPlanetEditor::WSimPlanetEditor(GSimulatorGameInterface& game_if, GSimPlanet& planet)
    : UIWindow(_("Edit Planet"), BLUE_WINDOW, COLOR_GRAY),
      planet(planet),
      game_if(game_if)
{
    init();
}

void
WSimPlanetEditor::init()
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));

    UIBaseWidget& tile1 = holder.add(new WSimPlanetPlanetTile(planet));
    UIBaseWidget& tile2 = holder.add(new WSimPlanetBaseTile(planet));

    tile1.observe(sig_edit);
    tile2.observe(sig_edit);

    buttons.add(holder.add(new UIButton(cm_Enter,  _("OK"), SDLK_RETURN)).addStop());
    buttons.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    buttons.add(holder.add(new UIButton(cm_No,     _("Del - No Planet"), SDLK_DELETE)).addStop());
    buttons.add(holder.add(new UISpacer()));
    buttons.add(holder.add(new UIButton(_("H"), 'h')));

    add(tile1);
    add(tile2);
    add(buttons);
    add(holder.add(new UIQuit(cm_Escape)));
    pack();

    /* q & d */
    for (UIBaseWidget* w = tile1.getFirstChild()->getFirstChild(); w != 0; w = w->getNextSibling())
        if (WSimPlanetEditButton* e = dynamic_cast<WSimPlanetEditButton*>(w))
            sig_edit.add(e, &WSimPlanetEditButton::update);
    for (UIBaseWidget* w = tile2.getFirstChild()->getFirstChild(); w != 0; w = w->getNextSibling())
        if (WSimPlanetEditButton* e = dynamic_cast<WSimPlanetEditButton*>(w))
            sig_edit.add(e, &WSimPlanetEditButton::update);
}

bool
WSimPlanetEditor::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case 'n':
            editName();
            update();
            return true;
         case 'o':
            editOwner();
            update();
            return true;
         case 'i':
            editId();
            update();
            return true;
         case 'f':
            editFCode();
            update();
            return true;
         case 'd':
            editDefense();
            update();
            return true;
         case 'p':
            editPopulation();
            update();
            return true;
         case 'l':
            ::editExperienceLevel(planet);
            update();
            break;
         case 'b':
            editBaseBeamTech();
            update();
            return true;
         case 'g':
            editBaseFighters();
            update();
            return true;
         case 's':
            editBaseDefense();
            update();
            return true;
         case 't':
            editBaseTorpTech();
            update();
            return true;
         case 'w':
            editBaseTorps();
            update();
            return true;
         case 'a':
            if (editSimulatorPlanetFunctions(planet)) {
                update();
            }
            return true;
         case 'h':
         case ss_Alt + 'h':
         case SDLK_F1:
            doHelp("pcc2:simplanet");
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
WSimPlanetEditor::update()
{
    sig_edit.raise();
}

/** Edit planet name. Since planet names can usually not be changed,
    we only offer a list of all names available in the game, and
    change the Id accordingly. */
void
WSimPlanetEditor::editName()
{
    UIStringList list;
    for (int i = 1; i <= NUM_PLANETS; ++i)
        list.add(i, game_if.getPlanetName(i));
    list.sortAlphabetically();

    UIStandardListbox box(0, 20, 20 * font_ems[FONT_NORMAL], list);
    box.scrollToKey(planet.getId());
    if (box.doStandardListbox(_("Set Planet Name"), "pcc2:simplanet")) {
        planet.setId(box.getKey());
        setPlanetName(game_if, planet);
    }
}

/** Edit planet owner. */
void
WSimPlanetEditor::editOwner()
{
    if (::editOwner(planet)) {
        if (planet.getBaseBeamTech() != 0) {
            int limit = config.MaximumFightersOnBase(planet.getOwner());
            if (planet.getBaseFighters() > limit)
                planet.setBaseFighters(limit);
            limit = config.MaximumDefenseOnBase(planet.getOwner());
            if (planet.getBaseDefense() > limit)
                planet.setBaseDefense(limit);
        }
    }
}

/** Edit planet Id. */
void
WSimPlanetEditor::editId()
{
    UIDecimalSelector select(0, 1, NUM_PLANETS, 10);
    select.setValue(planet.getId());
    if (select.doStandardDialog(_("Set Id"), _("Id:"), "pcc2:simplanet")) {
        planet.setId(select.getValue());
        setPlanetName(game_if, planet);
    }
}

/** Edit planet friendly code. */
void
WSimPlanetEditor::editFCode()
{
    /* build list of available codes */
    GFCodeList filtered_list;
    GFCodeList& full_list = GFCodeList::getInstance();

    for (GFCodeList::iterator i = full_list.begin(); i != full_list.end(); ++i) {
        if ((i->getFlags() & GFCode::fc_Planet) == 0)
            continue;
        filtered_list.addFCode(*i);
    }

    WFCodeWindow window(_("Set Friendly Code"), filtered_list);
    window.setFCode(planet.getFCode());

    ui_root->centerWidget(window);
    if (ui_root->runModal(window) == cm_Enter)
        planet.setFCode(window.getFCode());
}

/** Edit planetary defense. */
void
WSimPlanetEditor::editDefense()
{
    UIDecimalSelector select(0, 0, 600 /*FIXME: magic*/, 10);
    select.setValue(planet.getDefense());
    if (select.doStandardDialog(_("Set Defense"), _("Defense:"), "pcc2:simplanet"))
        planet.setDefense(select.getValue());
}

/** Edit planetary population. */
void
WSimPlanetEditor::editPopulation()
{
    /* predictions for instruction text */
    int defense = planet.getDefense();
    if (defense < 10)
        defense = 60;
    int clans;
    if (defense < 50)
        clans = defense;
    else
        clans = 50 + square(defense - 50);

    /* build a window */
    ObjectHolder h;

    UIWindow& window = h.add(new UIWindow(_("Edit Population"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    window.setLayout(UIVBoxLayout::instance5);

    UIInputLine& edit = h.add(new UIInputLine(6, 10, 0, UIInputLine::ilf_NumbersOnly | UIInputLine::ilf_Framed, 0));

    window.add(h.add(new UIMultilineStatic
                     (FONT_NORMAL,
                      format(_("Enter population of the planet. PCC2 will then "
                               "compute the maximum possible number of defense "
                               "posts supported by that population. For example, "
                               "%d clans support up to %d defense posts."),
                             clans, defense),
                      ui_screen->getSize().x / 2)));

    UIGroup& group_edit = h.add(new UIGroup(UIHBoxLayout::instance5));
    group_edit.add(edit);
    group_edit.add(h.add(new UIStaticText(_("clans"), UIStaticText::StaticTitle)));
    window.add(group_edit);

    UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
    group_buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    group_buttons.add(h.add(new UISpacer()));
    window.add(group_buttons);

    window.pack();
    ui_root->centerWidget(window);
    int rv = ui_root->runModal(window);
    if (rv == cm_Enter) {
        int new_clans;
        if (stringValue(edit.getText(), new_clans)) {
            // FIXME: this duplicates the formula from
            // getMaxStructuresFormula, but this avoids >>100k
            // dependency in PlayVCR
            int new_defense;
            if (new_clans <= 50)
                new_defense = new_clans;
            else
                new_defense = 50 + int(std::sqrt(double(new_clans - 50)) + 0.5);
            planet.setDefense(new_defense);
        }
    }
}

/** Edit starbase beam tech. */
void
WSimPlanetEditor::editBaseBeamTech()
{
    UIStringList list;
    list.add(0, _("0 - No base"));
    for (int i = 1; i <= NUM_BEAMS; ++i)
        list.add(i, format(_("%X - %s"), i, getBeam(i).getName()));

    UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    box.scrollToKey(planet.getBaseBeamTech());
    if (box.doStandardListbox(_("Set Starbase Beam Tech"), "pcc2:simplanet"))
        planet.setBaseBeamTech(box.getKey());
}

/** Edit number of starbase fighters. */
void
WSimPlanetEditor::editBaseFighters()
{
    if (planet.getBaseBeamTech() != 0) {
        UIDecimalSelector select(0, 0, config.MaximumFightersOnBase(planet.getOwner()), 10);
        select.setValue(planet.getBaseFighters());
        if (select.doStandardDialog(_("Set Base Fighters"), _("Fighters:"), "pcc2:simplanet"))
            planet.setBaseFighters(select.getValue());
    }
}

/** Edit number of starbase defense. */
void
WSimPlanetEditor::editBaseDefense()
{
    if (planet.getBaseBeamTech() != 0) {
        UIDecimalSelector select(0, 0, config.MaximumDefenseOnBase(planet.getOwner()), 10);
        select.setValue(planet.getBaseDefense());
        if (select.doStandardDialog(_("Set Base Defense"), _("Defense:"), "pcc2:simplanet"))
            planet.setBaseDefense(select.getValue());
    }
}

/** Edit starbase torpedo tech. */
void
WSimPlanetEditor::editBaseTorpTech()
{
    if (planet.getBaseBeamTech() != 0) {
        UIStringList list;
        for (int i = 1; i <= NUM_BEAMS; ++i)
            list.add(i, format(_("%X - %s"), i, getTorp(i).getName()));

        UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
        box.scrollToKey(planet.getBaseTorpTech());
        if (box.doStandardListbox(_("Set Starbase Torp Tech"), "pcc2:simplanet"))
            planet.setBaseTorpTech(box.getKey());
    }
}

/** Edit starbase torps. */
void
WSimPlanetEditor::editBaseTorps()
{
    if (planet.getBaseBeamTech() != 0) {
        editSimulatorPlanetTorps(planet);
    }
}

void
editSimulatorPlanet(GSimulatorGameInterface& game_if, GSimState& state)
{
    GSimPlanet edit;
    if (state.hasPlanet()) {
        edit = state.getPlanet();
    }
    setPlanetName(game_if, edit);

    WSimPlanetEditor editor(game_if, edit);
    ui_root->centerWidget(editor);
    int rv = ui_root->runModal(editor);

    switch (rv) {
     case cm_Enter:
        state.addPlanet();
        state.getPlanet() = edit;
        break;
     case cm_No:
        state.removePlanet();
        break;
    }
}
