/**
  *  \file client/widgets/billdisplay.cc
  *  \brief WBillDisplay and WBillTotalDisplay widgets
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/billdisplay.h"
#include "gfx/gfxcontext.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "cpluslib/format.h"
#include "util/misc.h"
#include "client/billexport.h"

/* PCC 1.x:
   - 500 pixels total
   - 25 for "N x"                2 em
   - 5 blank                   0.4 em
   - 270 for item name          20 em
   - 50 for $, T, D, M each      4 em
                              --------
                               ~39 em */

static void
showValue(GfxContext& ctx, int x, int y, int32_t value)
{
    if (value == 0) {
        ctx.setColor(UIColor::tc_Faded);
        outText(ctx, x, y, UTF_FIGURE_DASH);
    } else {
        ctx.setColor(UIColor::tc_Static);
        outText(ctx, x, y, numToString(value));
    }
}


/****************************** WBillDisplay *****************************/

/** Constructor.
    \param id Widget id
    \param nlines Number of lines for this widget
    \param is_list true to exhibit list behaviour, false to exhibit static behaviour
    \param bill Bill to display. Lifetime must be at least the lifetime of the widget */
WBillDisplay::WBillDisplay(int id, int nlines, bool is_list, const WBill& bill)
    : UIListbox(id, nlines, font_ems[FONT_NORMAL] * 39 + (is_list ? 5 : 0), bill.size()),
      bill(bill),
      is_list(is_list)
{
    setHeaderHeight(font_heights[FONT_NORMAL]);
    if (!is_list)
        setState(st_Disabled, true);
}

void
WBillDisplay::drawContent(GfxCanvas& can)
{
    const int x = getExtent().x;
    const int y = getExtent().y;
    const int m = font_ems[FONT_NORMAL];

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);
    outText(ctx, x, y, _("Item"));
    ctx.setTextAlign(2, 0);
    outText(ctx, x + 27*m, y, _("mc"));
    outText(ctx, x + 31*m, y, _("Tri"));
    outText(ctx, x + 35*m, y, _("Dur"));
    outText(ctx, x + 39*m, y, _("Mol"));
    drawHLine(ctx, x, y + font_heights[FONT_NORMAL]-2, x + 23*m);
    for (int i = 0; i < 4; ++i)
        drawHLine(ctx, x + (24+4*i)*m, y + font_heights[FONT_NORMAL]-2, x + (27+4*i)*m);

    can.update(GfxRect(x, y, getExtent().w, font_heights[FONT_NORMAL]));

    UIListbox::drawContent(can);
}

void
WBillDisplay::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    int m = font_ems[FONT_NORMAL];
    int item = getTop() + from;
    for (int line = from; line < to; ++line) {
        GfxRect area;
        prepareEntry(ctx, is_list ? ColorStyle : CursorlessColorStyle, line, &area);
        if (item < int(bill.size())) {
            ctx.setTextAlign(0, 0);
            outTextF(ctx, area.x + 3*m, area.y, m*20, bill[item]->item_name);
            ctx.setTextAlign(2, 0);
            outText(ctx, area.x + 3*m,  area.y, format("%d %s ", bill[item]->display_mult, UTF_TIMES));
            showValue(ctx, area.x + 27*m, area.y, bill[item]->item_cost.get(el_Money) + bill[item]->item_cost.get(el_Supplies));
            showValue(ctx, area.x + 31*m, area.y, bill[item]->item_cost.get(el_Tritanium));
            showValue(ctx, area.x + 35*m, area.y, bill[item]->item_cost.get(el_Duranium));
            showValue(ctx, area.x + 39*m, area.y, bill[item]->item_cost.get(el_Molybdenum));
        }
        ++item;
    }
}

void
WBillDisplay::update()
{
    setTop(0);
    setCurrent(0);
    setSize(bill.size());
    drawWidget();
}

void
WBillDisplay::doExport()
{
    doBillExport(bill);
}

/*************************** WBillTotalDisplay ***************************/

/** Constructor for single-line version.
    \param bill Bill to display */
WBillTotalDisplay::WBillTotalDisplay(const WBill& bill)
    : bill(bill),
      have(),
      has_compare(false)
{ }

/** Constructor for two-line version.
    \param have Available amounts
    \param bill Bill to display */
WBillTotalDisplay::WBillTotalDisplay(const GCost& have, const WBill& bill)
    : bill(bill),
      have(have),
      has_compare(true)
{ }

void
WBillTotalDisplay::drawContent(GfxCanvas& can)
{
    int x = getExtent().x;
    int y = getExtent().y;
    int h = font_heights[FONT_NORMAL];
    int m = font_ems[FONT_NORMAL];

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    drawBackground(ctx, getExtent());

    // Divider
    y++;
    drawHLine(ctx, x, y, x + 23*m);
    for (int i = 0; i < 4; ++i)
        drawHLine(ctx, x + (24+4*i)*m, y, x + (27+4*i)*m);
    y++;

    // Compute totals
    GCost total;
    for (WBill::const_iterator i = bill.begin(); i != bill.end(); ++i)
        total += (*i)->item_cost;
    outText(ctx, x, y, _("Total:"));

    // Display it
    if (!has_compare) {
        // Simple version
        ctx.setTextAlign(2, 0);
        showValue(ctx, x + 27*m, y, total.get(el_Money) + total.get(el_Supplies));
        showValue(ctx, x + 31*m, y, total.get(el_Tritanium));
        showValue(ctx, x + 35*m, y, total.get(el_Duranium));
        showValue(ctx, x + 39*m, y, total.get(el_Molybdenum));
    } else {
        // With comparison
        outText(ctx, x, y+h, _("Remaining:"));
        ctx.setTextAlign(2, 0);

        // MC/Supplies
        int32_t haveMC  = have.get(el_Money);
        int32_t haveSup = have.get(el_Supplies);
        int32_t needMC  = total.get(el_Money);
        int32_t needSup = total.get(el_Supplies);
        if (haveMC+haveSup < needMC+needSup) {
            // There's a shortage of mc+sup, so we need more supplies
            ctx.setColor(UIColor::tc_Red);
            outText(ctx, x + 27*m, y+h, format(_("(sup) %d"), numToString(haveSup+haveMC - needSup - needMC)));
        } else if (haveSup < needSup) {
            // Enough money, but not enough supplies
            ctx.setColor(UIColor::tc_Red);
            outText(ctx, x + 27*m, y+h, format(_("(sup) %d"), numToString(haveSup - needSup)));
        } else {
            // Enough money and supplies, but we may need to sell supplies
            ctx.setColor(UIColor::tc_Green);
            if (haveMC < needMC) {
                outText(ctx, x + 27*m, y+h, format(_("(sup) %d"), numToString(haveSup+haveMC - needSup - needMC)));
            } else {
                outText(ctx, x + 27*m, y+h, format(_("(mc) %d"), numToString(haveMC - needMC)));
            }
        }
        outText(ctx, x + 27*m, y, numToString(total.get(el_Money) + total.get(el_Supplies)));

        // Minerals
        for (int i = 0; i < 3; ++i) {
            GCargoType ct = GCargoType(el_Tritanium+i);
            int32_t delta = have.get(ct) - total.get(ct);
            ctx.setColor(delta >= 0 ? UIColor::tc_Green : UIColor::tc_Red);
            outText(ctx, x + (31+4*i)*m, y, numToString(total.get(ct)));
            outText(ctx, x + (31+4*i)*m, y+h, numToString(delta));
        }
    }

    can.update(getExtent());
}

void
WBillTotalDisplay::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint pt = font_sizes[FONT_NORMAL];
    pt.x *= 39;
    if (has_compare)
        pt.y *= 2;
    pt.y += 2;
    
    info.pref_size = info.min_size = pt;
    info.flags = 0;
}

void
WBillTotalDisplay::setAvailableResources(const GCost& have)
{
    this->have = have;
    drawWidget();
}

void
WBillTotalDisplay::update()
{
    drawWidget();
}
