/**
  *  \file game/undo.cc
  *  \brief Undo Validation for Planets
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/undo.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/cargotype.h"
#include "game/spec.h"
#include "game/config.h"
#include "game/global.h"

/** Default constructor. Makes a blank object. */
GUndoInformation::GUndoInformation()
{
    clear();
}

/** Compute undo information for one place in space.
    \param trn Turn to work on
    \param pid Planet to work on. All ships in orbit will be considered. */
void
GUndoInformation::computeInformation(const GGameTurn& trn, const int pid)
{
    clear();

    /* fetch planets */
    const GPlanet& dat = trn.getCurrentUniverse().getPlanet(pid);
    const GPlanet& dis = trn.getPreviousUniverse().getPlanet(pid);
    if (!dat.isPlayable(GObject::Playable))
        return;

    /* we will count resources used. That is, if something consumes
       supplies, increase supply_diffs. */
    int32_t supply_diffs = 0;

    /* we will count units built. */
    int32_t fighter_diffs = 0;
    int32_t torp_diffs[NUM_TORPS] = { 0 };

    /* count planet: supply diffs only */
    supply_diffs += dat.getStructures(str_Factories) - dis.getStructures(str_Factories);
    supply_diffs += dat.getStructures(str_Mines)     - dis.getStructures(str_Mines);
    supply_diffs += dat.getStructures(str_Defense)   - dis.getStructures(str_Defense);

    if (dat.isBuildingBase() && !dis.isBuildingBase())
        supply_diffs += config.StarbaseCost(dat.getOwner()).get(el_Supplies);

    /* count base: ammo diffs */
    if (dat.hasBase() && dis.hasBase()) {
        fighter_diffs += dat.getCargo(el_Fighters) - dis.getCargo(el_Fighters);
        for (int i = 1; i <= NUM_TORPS; ++i)
            torp_diffs[i-1] = dat.getCargo(getCargoTypeFromTorpType(i)) - dis.getCargo(getCargoTypeFromTorpType(i));
    }

    /* count ships: ammo and supply diffs */
    const GPlayedShipType& ty = trn.getCurrentUniverse().ty_played_ships;
    for (int sid = ty.findNextIndex(0); sid != 0; sid = ty.findNextIndex(sid)) {
        const GShip& sdat = trn.getCurrentUniverse().getShip(sid);
        if (sdat.getOwner() == dat.getOwner() && sdat.getPos() == dat.getPos()) {
            const GShip& sdis = trn.getPreviousUniverse().getShip(sid);
            supply_diffs += sdat.getCargo(el_Supplies) - sdis.getCargo(el_Supplies);
            if (sdat.getNumBays() > 0)
                fighter_diffs += sdat.getAmmo() - sdis.getAmmo();
            else if (sdat.getNumTorpLaunchers() > 0)
                torp_diffs[sdat.getTorpType()-1] +=
                    sdat.getAmmo() - sdis.getAmmo();

            if (sdat.isTransporterActive(GShip::t_Transfer)) {
                supply_diffs += sdat.getTransporter(GShip::t_Transfer).supplies;
            }
            if (sdis.isTransporterActive(GShip::t_Transfer)) {
                supply_diffs -= sdis.getTransporter(GShip::t_Transfer).supplies;
            }
            if (sdat.isTransporterActive(GShip::t_Unload)) {
                supply_diffs += sdat.getTransporter(GShip::t_Unload).supplies;
            }
            if (sdis.isTransporterActive(GShip::t_Unload)) {
                supply_diffs -= sdis.getTransporter(GShip::t_Unload).supplies;
            }
        }
    }

    /* if we built fighters, include them in the supply count */
    if (fighter_diffs > 0)
        supply_diffs += fighter_diffs * config.BaseFighterCost(dat.getOwner()).get(el_Supplies);

    /* produce output for ammo diffs */
    if (fighter_diffs > 0)
        fighters_to_sell = fighter_diffs;
    for (int i = 0; i < NUM_TORPS; ++i)
        if (torp_diffs[i] > 0)
            torps_to_sell[i] = torp_diffs[i];

    /* produce output for supply diffs: we started the turn with X
       supplies, used Y and sold Z, so we end up at X' = X-Y-Z, or
       Z=X-Y-X'. */
    supplies_to_buy = dis.getCargo(el_Supplies) - supply_diffs - dat.getCargo(el_Supplies);
    if (supplies_to_buy < 0)
        supplies_to_buy = 0;

    /* produce output for tech levels */
    if (dat.hasBase() && dis.hasBase()) {
        int level;

        /* hulls */
        level = dis.getBaseTechLevel(tl_Hull);
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
            if (dat.getBaseHullStoreSlot(i) > dis.getBaseHullStoreSlot(i)) {
                int tech = getHull(getTruehull(dat.getOwner(), i)).getTech();
                if (tech > level)
                    level = tech;
            }
        }
        minimum_tech_levels[tl_Hull] = level;
        
        /* engines */
        level = dis.getBaseTechLevel(tl_Engine);
        for (int i = 1; i <= NUM_ENGINES; ++i) {
            if (dat.getBaseEngineStore(i) > dis.getBaseEngineStore(i)) {
                int tech = getEngine(i).getTech();
                if (tech > level)
                    level = tech;
            }
        }
        minimum_tech_levels[tl_Engine] = level;

        /* beams */
        level = dis.getBaseTechLevel(tl_Beam);
        for (int i = 1; i <= NUM_BEAMS; ++i) {
            if (dat.getBaseBeamStore(i) > dis.getBaseBeamStore(i)) {
                int tech = getBeam(i).getTech();
                if (tech > level)
                    level = tech;
            }
        }
        minimum_tech_levels[tl_Beam] = level;

        /* torpedoes */
        level = dis.getBaseTechLevel(tl_Torpedo);
        for (int i = 1; i <= NUM_TORPS; ++i) {
            if (dat.getBaseLauncherStore(i) > dis.getBaseLauncherStore(i) || torp_diffs[i-1] > 0) {
                int tech = getTorp(i).getTech();
                if (tech > level)
                    level = tech;
            }
        }
        minimum_tech_levels[tl_Torpedo] = level;
    }
}

/** Clear this object. */
void
GUndoInformation::clear()
{
    for (int i = 0; i < NUM_TORPS; ++i)
        torps_to_sell[i] = 0;
    fighters_to_sell = 0;
    supplies_to_buy = 0;
    for (int i = 0; i < 4; ++i)
        minimum_tech_levels[i] = 0;
}

/** Get amount of ammunition that can be sold.
    \param type Cargo type to query
    \return amount that can be sold */
int32_t
GUndoInformation::getAmmoAllowedToSell(GCargoType type) const
{
    if (type == el_Fighters)
        return fighters_to_sell;
    else if (isCargoTorpedo(type))
        return torps_to_sell[getTorpTypeFromCargoType(type) - 1];
    else
        return 0;
}

/** Get number of supplies that can be bought. */
int32_t
GUndoInformation::getSuppliesAllowedToBuy() const
{
    return supplies_to_buy;
}

/** Get minimum tech level permitted on base. */
int
GUndoInformation::getMinimumTech(TTechLevel level) const
{
    return minimum_tech_levels[level];
}
