/**
  *  \file int/if/vcrif.cc
  *  \brief Interpreter Game Interface: VCRs
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/racename.h"
#include "game/spec.h"
#include "game/vcrdb.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "int/if/vcrif.h"
#include "int/interror.h"
#include "int/value.h"
#include "io/io.h"
#include "util/nls.h"
#include "util/recode.h"
#include "int/propenum.h"
#include "game/classicvcr.h"
#include "int/compound.h"

/* @q Vcr(uid:Int):Obj (Function, Context)
   Access properties of combat recordings.
   Use as
   | ForEach Vcr Do ...
   or
   | With Vcr(n) Do ...
   @see int:index:group:combatproperty|Combat Properties */
IntValue*
IFVcrGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!haveDisplayedTurn())
        return 0;
    if (!checkIntArg(id, args.getNext(), 1, getDisplayedTurn().getVcrs().getNumBattles()))
        return 0;
    return new IntVcrContext(id);
}

IntContext*
IFVcrMake()
{
    if (!haveDisplayedTurn())
        return 0;

    GVcrDatabase& db = getDisplayedTurn().getVcrs();
    if (db.getNumBattles() == 0)
        return 0;

    return new IntVcrContext(1);
}

int32_t
IFVcrDim(int32_t dim)
{
    if (!haveDisplayedTurn()) {
        return 0;
    } else if (dim == 0) {
        return 1;
    } else {
        return getDisplayedTurn().getVcrs().getNumBattles()+1;
    }
}

/***************************** IntVcrContext *****************************/

namespace { enum { VcrProperty, LeftProperty, RightProperty }; }

static const TIntMapping vcr_mapping[] = {
    { "ALGORITHM",                ivpAlgorithm,     VcrProperty,   thString },
    { "CAPABILITIES",             ivpFlags,         VcrProperty,   thInt },
    { "LEFT",                     ivsNameFull,      LeftProperty,  thString },
    { "LEFT.AUX",                 ivsAuxName,       LeftProperty,  thString },
    { "LEFT.AUX$",                ivsAuxId,         LeftProperty,  thInt },
    { "LEFT.AUX.AMMO",            ivsAuxAmmo,       LeftProperty,  thInt },
    { "LEFT.AUX.COUNT",           ivsAuxCount,      LeftProperty,  thInt },
    { "LEFT.AUX.SHORT",           ivsAuxShort,      LeftProperty,  thString },
    { "LEFT.BEAM",                ivsBeamName,      LeftProperty,  thString },
    { "LEFT.BEAM$",               ivsBeamId,        LeftProperty,  thInt },
    { "LEFT.BEAM.COUNT",          ivsBeamCount,     LeftProperty,  thInt },
    { "LEFT.BEAM.SHORT",          ivsBeamShort,     LeftProperty,  thString },
    { "LEFT.CREW",                ivsCrew,          LeftProperty,  thInt },
    { "LEFT.CREW$",               ivsCrewRaw,       LeftProperty,  thInt },
    { "LEFT.DAMAGE",              ivsDamage,        LeftProperty,  thInt },
    { "LEFT.FIGHTER.BAYS",        ivsFighterBays,   LeftProperty,  thInt },
    { "LEFT.FIGHTER.COUNT",       ivsFighterCount,  LeftProperty,  thInt },
    { "LEFT.HULL",                ivsHullName,      LeftProperty,  thString },
    { "LEFT.HULL$",               ivsHullId,        LeftProperty,  thInt },
    { "LEFT.ID",                  ivsId,            LeftProperty,  thInt },
    { "LEFT.IMAGE",               ivsImage,         LeftProperty,  thInt },
    { "LEFT.LEVEL",               ivsLevel,         LeftProperty,  thInt },
    { "LEFT.MASS",                ivsMass,          LeftProperty,  thInt },
    { "LEFT.NAME",                ivsName,          LeftProperty,  thString },
    { "LEFT.OWNER",               ivsOwnerShort,    LeftProperty,  thString },
    { "LEFT.OWNER$",              ivsOwnerId,       LeftProperty,  thInt },
    { "LEFT.OWNER.ADJ",           ivsOwnerAdj,      LeftProperty,  thString },
    { "LEFT.SHIELD",              ivsShield,        LeftProperty,  thInt },
    { "LEFT.STATUS",              ivsStatus,        LeftProperty,  thString },
    { "LEFT.STATUS$",             ivsStatusRaw,     LeftProperty,  thString },
    { "LEFT.TORP",                ivsTorpName,      LeftProperty,  thString },
    { "LEFT.TORP$",               ivsTorpId,        LeftProperty,  thInt },
    { "LEFT.TORP.COUNT",          ivsTorpCount,     LeftProperty,  thInt },
    { "LEFT.TORP.LCOUNT",         ivsTorpLCount,    LeftProperty,  thInt },
    { "LEFT.TORP.SHORT",          ivsTorpShort,     LeftProperty,  thString },
    { "LEFT.TYPE",                ivsType,          LeftProperty,  thString },
    { "LEFT.TYPE.SHORT",          ivsTypeShort,     LeftProperty,  thString },
    { "MAGIC",                    ivpMagic,         VcrProperty,   thInt },
    { "NUMUNITS",                 ivpNumUnits,      VcrProperty,   thInt },
    { "RIGHT",                    ivsNameFull,      RightProperty, thString },
    { "RIGHT.AUX",                ivsAuxName,       RightProperty, thString },
    { "RIGHT.AUX$",               ivsAuxId,         RightProperty, thInt },
    { "RIGHT.AUX.AMMO",           ivsAuxAmmo,       RightProperty, thInt },
    { "RIGHT.AUX.COUNT",          ivsAuxCount,      RightProperty, thInt },
    { "RIGHT.AUX.SHORT",          ivsAuxShort,      RightProperty, thString },
    { "RIGHT.BEAM",               ivsBeamName,      RightProperty, thString },
    { "RIGHT.BEAM$",              ivsBeamId,        RightProperty, thInt },
    { "RIGHT.BEAM.COUNT",         ivsBeamCount,     RightProperty, thInt },
    { "RIGHT.BEAM.SHORT",         ivsBeamShort,     RightProperty, thString },
    { "RIGHT.CREW",               ivsCrew,          RightProperty, thInt },
    { "RIGHT.CREW$",              ivsCrewRaw,       RightProperty, thInt },
    { "RIGHT.DAMAGE",             ivsDamage,        RightProperty, thInt },
    { "RIGHT.FIGHTER.BAYS",       ivsFighterBays,   RightProperty, thInt },
    { "RIGHT.FIGHTER.COUNT",      ivsFighterCount,  RightProperty, thInt },
    { "RIGHT.HULL",               ivsHullName,      RightProperty, thString },
    { "RIGHT.HULL$",              ivsHullId,        RightProperty, thInt },
    { "RIGHT.ID",                 ivsId,            RightProperty, thInt },
    { "RIGHT.IMAGE",              ivsImage,         RightProperty, thInt },
    { "RIGHT.LEVEL",              ivsLevel,         RightProperty, thInt },
    { "RIGHT.MASS",               ivsMass,          RightProperty, thInt },
    { "RIGHT.NAME",               ivsName,          RightProperty, thString },
    { "RIGHT.OWNER",              ivsOwnerShort,    RightProperty, thString },
    { "RIGHT.OWNER$",             ivsOwnerId,       RightProperty, thInt },
    { "RIGHT.OWNER.ADJ",          ivsOwnerAdj,      RightProperty, thString },
    { "RIGHT.SHIELD",             ivsShield,        RightProperty, thInt },
    { "RIGHT.STATUS",             ivsStatus,        RightProperty, thString },
    { "RIGHT.STATUS$",            ivsStatusRaw,     RightProperty, thString },
    { "RIGHT.TORP",               ivsTorpName,      RightProperty, thString },
    { "RIGHT.TORP$",              ivsTorpId,        RightProperty, thInt },
    { "RIGHT.TORP.COUNT",         ivsTorpCount,     RightProperty, thInt },
    { "RIGHT.TORP.LCOUNT",        ivsTorpLCount,    RightProperty, thInt },
    { "RIGHT.TORP.SHORT",         ivsTorpShort,     RightProperty, thString },
    { "RIGHT.TYPE",               ivsType,          RightProperty, thString },
    { "RIGHT.TYPE.SHORT",         ivsTypeShort,     RightProperty, thString },
    { "SEED",                     ivpSeed,          VcrProperty,   thInt },
    { "TYPE$",                    ivpType,          VcrProperty,   thInt },
    { "UNIT",                     ivpUnits,         VcrProperty,   thArray },
};

IntVcrContext::IntVcrContext(int vcrnr)
    : vcrnr(vcrnr)
{ }

IntVcrContext::~IntVcrContext()
{ }

// IntContext:
int32_t
IntVcrContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, vcr_mapping, countof(vcr_mapping));
}

void
IntVcrContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntVcrContext::get(int32_t index)
{
    /* Attempt to fetch VCR entry */
    if (!haveDisplayedTurn())
        return 0;

    GVcrDatabase& db = getDisplayedTurn().getVcrs();
    if (vcrnr <= 0 || vcrnr > db.getNumBattles())
        return 0;

    /* We have a VCR, and we have a valid index, so this will work out */
    GVcrEntry& vcr = db.getBattle(vcrnr-1);
    switch (vcr_mapping[index].domain) {
     case VcrProperty:
        return getVcrProperty(vcr, IntVcrProperty(vcr_mapping[index].index));
     case LeftProperty:
        return getVcrSideProperty(vcr, 0, IntVcrSideProperty(vcr_mapping[index].index));
     case RightProperty:
        return getVcrSideProperty(vcr, 1, IntVcrSideProperty(vcr_mapping[index].index));
     default:
        return 0;
    }
}

bool
IntVcrContext::next()
{
    if (haveDisplayedTurn() && vcrnr < getDisplayedTurn().getVcrs().getNumBattles()) {
        ++vcrnr;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntVcrContext::clone()
{
    return new IntVcrContext(vcrnr);
}

GObject*
IntVcrContext::getObject()
{
    return 0;
}

string_t
IntVcrContext::toString(bool /*readable*/)
{
    return format("Vcr(%d)", vcrnr);
}

void
IntVcrContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notAssignable();
}

void
IntVcrContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(vcr_mapping, countof(vcr_mapping));
}

/********************************* Sides *********************************/

namespace {
    class IntVcrSideArray : public IntIndexableValue {
     public:
        IntVcrSideArray(GVcrEntry& e);

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        int32_t getDimension(int32_t dim);
        IntContext* makeFirstContext();

        // IntValue:
        IntVcrSideArray* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
     private:
        GVcrEntry& entry;
    };

    class IntVcrSideContext : public IntContext {
     public:
        IntVcrSideContext(GVcrEntry& entry, int side);
        ~IntVcrSideContext();

        // IntContext:
        int32_t   lookup(const IntNameQuery& name);
        void      set(int32_t index, IntValue* value);
        IntValue* get(int32_t index);
        bool      next();
        IntContext* clone();
        GObject*  getObject();
        string_t  toString(bool readable);
        void      store(IntTagNode& sv, Stream& aux);
        void      enumProperties(IntPropertyAcceptor& acceptor);

     private:
        GVcrEntry& entry;
        int side;
    };
}

IntVcrSideArray::IntVcrSideArray(GVcrEntry& e)
    : entry(e)
{ }

IntValue*
IntVcrSideArray::get(IntArgBlock& args)
{
    // Validate args
    int32_t i;
    args.checkArgumentCount(1);
    if (!checkIntArg(i, args.getNext(), 1, entry.getNumObjects())) {
        return 0;
    }

    // Get result
    return new IntVcrSideContext(entry, i-1);
}

void
IntVcrSideArray::set(IntArgBlock&, IntValue*)
{
    throw IntError::notAssignable();
}

int32_t
IntVcrSideArray::getDimension(int32_t dim)
{
    return dim == 0 ? 1 : entry.getNumObjects()+1;
}

IntContext*
IntVcrSideArray::makeFirstContext()
{
    if (entry.getNumObjects() > 0) {
        return new IntVcrSideContext(entry, 0);
    } else {
        return 0;
    }
}

IntVcrSideArray*
IntVcrSideArray::clone()
{
    return new IntVcrSideArray(entry);
}

string_t
IntVcrSideArray::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntVcrSideArray::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

namespace { enum { SideProperty, SideId }; }

static const TIntMapping side_mapping[] = {
    { "AUX",                 ivsAuxName,       SideProperty,  thString },
    { "AUX$",                ivsAuxId,         SideProperty,  thInt },
    { "AUX.AMMO",            ivsAuxAmmo,       SideProperty,  thInt },
    { "AUX.COUNT",           ivsAuxCount,      SideProperty,  thInt },
    { "AUX.SHORT",           ivsAuxShort,      SideProperty,  thString },
    { "BEAM",                ivsBeamName,      SideProperty,  thString },
    { "BEAM$",               ivsBeamId,        SideProperty,  thInt },
    { "BEAM.COUNT",          ivsBeamCount,     SideProperty,  thInt },
    { "BEAM.SHORT",          ivsBeamShort,     SideProperty,  thString },
    { "CONFIG.BEAMCHARGERATE", ivsBeamChargeRate, SideProperty, thInt },
    { "CONFIG.BEAMKILLRATE", ivsBeamKillRate,  SideProperty,  thInt },
    { "CONFIG.CREWDEFENSERATE", ivsCrewDefenseRate, SideProperty, thInt },
    { "CONFIG.TORPCHARGERATE", ivsTorpChargeRate, SideProperty, thInt },
    { "CONFIG.TORPMISSRATE", ivsTorpMissRate,  SideProperty,  thInt },
    { "CREW",                ivsCrew,          SideProperty,  thInt },
    { "CREW$",               ivsCrewRaw,       SideProperty,  thInt },
    { "DAMAGE",              ivsDamage,        SideProperty,  thInt },
    { "FIGHTER.BAYS",        ivsFighterBays,   SideProperty,  thInt },
    { "FIGHTER.COUNT",       ivsFighterCount,  SideProperty,  thInt },
    { "HULL",                ivsHullName,      SideProperty,  thString },
    { "HULL$",               ivsHullId,        SideProperty,  thInt },
    { "ID",                  ivsId,            SideProperty,  thInt },
    { "IMAGE",               ivsImage,         SideProperty,  thInt },
    { "INDEX",               0,                SideId,        thInt },
    { "ISPLANET",            ivsIsPlanet,      SideProperty,  thBool },
    { "LEVEL",               ivsLevel,         SideProperty,  thInt },
    { "MASS",                ivsMass,          SideProperty,  thInt },
    { "NAME",                ivsName,          SideProperty,  thString },
    { "NAME.FULL",           ivsNameFull,      SideProperty,  thString },
    { "OWNER",               ivsOwnerShort,    SideProperty,  thString },
    { "OWNER$",              ivsOwnerId,       SideProperty,  thInt },
    { "OWNER.ADJ",           ivsOwnerAdj,      SideProperty,  thString },
    { "SHIELD",              ivsShield,        SideProperty,  thInt },
    { "STATUS",              ivsStatus,        SideProperty,  thString },
    { "STATUS$",             ivsStatusRaw,     SideProperty,  thString },
    { "TORP",                ivsTorpName,      SideProperty,  thString },
    { "TORP$",               ivsTorpId,        SideProperty,  thInt },
    { "TORP.COUNT",          ivsTorpCount,     SideProperty,  thInt },
    { "TORP.LCOUNT",         ivsTorpLCount,    SideProperty,  thInt },
    { "TORP.SHORT",          ivsTorpShort,     SideProperty,  thString },
    { "TYPE",                ivsType,          SideProperty,  thString },
    { "TYPE.SHORT",          ivsTypeShort,     SideProperty,  thString },
};


IntVcrSideContext::IntVcrSideContext(GVcrEntry& entry, int side)
    : entry(entry),
      side(side)
{ }

IntVcrSideContext::~IntVcrSideContext()
{ }

int32_t
IntVcrSideContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, side_mapping, countof(side_mapping));
}

void
IntVcrSideContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntVcrSideContext::get(int32_t index)
{
    switch (side_mapping[index].domain) {
     case SideProperty:
        return getVcrSideProperty(entry, side, IntVcrSideProperty(side_mapping[index].index));
     case SideId:
        /* @q Index:Int (Combat Participant Property)
           Position of this unit in the fight.
           This is the index into the fight's {Unit (Combat Property)|Unit} array. */
        return makeIntValue(side+1);
     default:
        return 0;
    }
}

bool
IntVcrSideContext::next()
{
    if (side+1 < entry.getNumObjects()) {
        ++side;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntVcrSideContext::clone()
{
    return new IntVcrSideContext(entry, side);
}

GObject*
IntVcrSideContext::getObject()
{
    return 0;
}

string_t
IntVcrSideContext::toString(bool /*readable*/)
{
    return "#<vcr-object>";
}

void
IntVcrSideContext::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

void
IntVcrSideContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(side_mapping, countof(side_mapping));
}

/****************************** Name Lookup ******************************/

/** Convert status flags to script value.
    \param status Status value
    \param side Side to query */
static IntValue*
makeVcrResult(GVcrEntry& e, int side)
{
    if (e.getPlayability() != GVcrEntry::IsPlayable) {
        return makeStringValue("Invalid");
    } else {
        /* This originally distinguished between Stalemate, Timeout and Disabled,
           which can be derived from the GClassicVcrEntry interface. Ultimately,
           all three cases are a Stalemate, so we use that. How to detect a
           stalemate? We declare a stalemate if ALL units in a fight survive.
           If anyone dies or gets captured, it is no longer a stalemate. */
        int status = e.getOutcome(side);
        if (status == 0) {
            for (int i = 0, n = e.getNumObjects(); i < n; ++i) {
                if (e.getOutcome(i) != 0) {
                    return makeStringValue("Survived");
                }
            }
            return makeStringValue("Stalemate");
        } else if (status < 0) {
            return makeStringValue("Exploded");
        } else {
            return makeStringValue("Captured");
        }
    }
}

/** Get property of a VCR side. Note that the outcome is a property of
    a VCR side. Therefore, this function may need to invoke the VCR
    player. For this reason, the GVcrEntry cannot be const, because it
    will be updated with the computed result.

    There is no longer any protection against this calling the (costly) VCR.
    Therefore, the server module must make sure to never call this for
    ivsStatus / ivsStatusRaw.

    \param vcr The record.
    \param side Side to query (0=left, 1=right)
    \param ivs Property to get */
IntValue*
getVcrSideProperty(GVcrEntry& vcr, int side, IntVcrSideProperty ivs)
{
    /* Note: this implements the owner properties directly instead of relying on
       getPlayerProperty() like most others. This makes it possible to use VCR
       properties without having a turn loaded. */
    /* FIXME: this does not implement PlanetsHaveTubes; for planets with tubes,
       it always returns the fighters. */
    const GVcrObject& o = vcr.getObject(side, false);
    switch (ivs) {
     case ivsAuxAmmo:
        /* Aux.Ammo documented in getShipProperty */
        /* @q Left.Aux.Ammo:Int (Combat Property), Right.Aux.Ammo:Int (Combat Property)
           Number of fighters or torpedoes on this unit. */
        if (o.getBayCount() > 0) {
            return makeIntValue(o.getFighterCount());
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpCount());
        } else {
            return makeIntValue(0);
        }
     case ivsAuxCount:
        /* Aux.Count documented in getShipProperty */
        /* @q Left.Aux.Count:Int (Combat Property), Right.Aux.Count:Int (Combat Property)
           Number of fighter bays or torpedo launchers on this unit. */
        if (o.getBayCount() > 0) {
            return makeIntValue(o.getBayCount());
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpLauncherCount());
        } else {
            return 0;
        }
     case ivsAuxId:
        /* Aux$ documented in getShipProperty */
        /* @q Left.Aux$:Int (Combat Property), Right.Aux$:Int (Combat Property)
           Type of secondary weapon.
           - 1..10 for torpedoes
           - 11 for fighters
           - EMPTY if no secondary weapon. */
        if (o.getBayCount() > 0) {
            return makeIntValue(11);
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpType());
        } else {
            return 0;
        }
     case ivsAuxName:
        /* Aux documented in getShipProperty */
        /* @q Left.Aux:Str (Combat Property), Right.Aux:Str (Combat Property)
           Secondary weapon type, full name.
           Either a torpedo system name, "Fighters", or EMPTY. */
        if (o.getBayCount() > 0) {
            return makeStringValue("Fighters");
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getName());
        } else {
            return 0;
        }
     case ivsAuxShort:
        /* Aux.Short documented in getShipProperty */
        /* @q Left.Aux.Short:Str (Combat Property), Right.Aux.Short:Str (Combat Property)
           Secondary weapon type, short name.
           @see Left.Aux (Combat Property) */
        if (o.getBayCount() > 0) {
            return makeStringValue("Ftr");
        } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getShortName());
        } else {
            return 0;
        }
     case ivsFighterBays:
        /* Fighter.Bays documented in getShipProperty */
        /* @q Left.Fighter.Bays:Int (Combat Property), Right.Fighter.Bays:Int (Combat Property)
           Number of fighter bays. */
        return makeIntValue(o.getBayCount());
     case ivsFighterCount:
        /* Fighter.Count documented in getShipProperty */
        /* @q Left.Fighter.Count:Int (Combat Property), Right.Fighter.Count:Int (Combat Property)
           Number of fighters. */
        return makeIntValue(o.getBayCount() > 0 ? o.getFighterCount() : 0);
     case ivsTorpId:
        /* Torp$ documented in getShipProperty */
        /* @q Left.Torp$:Int (Combat Property), Right.Torp$:Int (Combat Property)
           Torpedo type. */
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpType());
        } else {
            return 0;
        }
     case ivsTorpCount:
        /* Torp.Count documented in getShipProperty */
        /* @q Left.Torp.Count:Int (Combat Property), Right.Torp.Count:Int (Combat Property)
           Number of torpedoes. 0 if the ship has no torpedoes. */
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpCount());
        } else {
            return makeIntValue(0);
        }

     case ivsTorpLCount:
        /* Torp.LCount documented in getShipProperty */
        /* @q Left.Torp.LCount:Int (Combat Property), Right.Torp.LCount:Int (Combat Property)
           Number of torpedo launchers on this ship. */
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeIntValue(o.getTorpLauncherCount());
        } else {
            return makeIntValue(0);
        }
     case ivsTorpShort:
        /* Torp.Short documented in getShipProperty */
        /* @q Left.Torp.Short:Str (Combat Property), Right.Torp.Short:Str (Combat Property)
           Torpedo type, short name. */
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getShortName());
        } else {
            return 0;
        }
     case ivsTorpName:
        /* Torp documented in getShipProperty */
        /* @q Left.Torp:Str (Combat Property), Right.Torp:Str (Combat Property)
           Torpedo type, full name. */
        if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.getTorpType()).getName());
        } else {
            return 0;
        }
     case ivsBeamCount:
        /* Beam.Count documented in getShipProperty */
        /* @q Left.Beam.Count:Int (Combat Property), Right.Beam.Count:Int (Combat Property)
           Number of beams. */
        return makeIntValue(o.getBeamCount());
     case ivsBeamId:
        /* Beam$ documented in getShipProperty */
        /* @q Left.Beam$:Int (Combat Property), Right.Beam$:Int (Combat Property)
           Beam type. 0 if none, EMPTY if not known. */
        return makeIntValue(o.getBeamType());
     case ivsBeamName:
        /* Beam documented in getShipProperty */
        /* @q Left.Beam:Str (Combat Property), Right.Beam:Str (Combat Property)
           Beam type, full name. */
        if (o.getBeamType() > 0 && o.getBeamType() <= NUM_BEAMS) {
            return makeStringValue(::getBeam(o.getBeamType()).getName());
        } else {
            return 0;
        }
     case ivsBeamShort:
        /* Beam.Short documented in getShipProperty */
        /* @q Left.Beam.Short:Str (Combat Property), Right.Beam.Short:Str (Combat Property)
           Beam type, short name. */
        if (o.getBeamType() > 0 && o.getBeamType() <= NUM_BEAMS) {
            return makeStringValue(::getBeam(o.getBeamType()).getShortName());
        } else {
            return 0;
        }
     case ivsCrew:
        /* @q Crew:Int (Combat Participant Property)
           @q Left.Crew:Int (Combat Property), Right.Crew:Int (Combat Property)
           Crew on this ship. EMPTY if this is a planet. */
        if (o.isPlanet()) {
            return 0;
        } else {
            return makeIntValue(o.getCrew());
        }
     case ivsCrewRaw:
        /* @q Crew$:Int (Combat Participant Property)
           @q Left.Crew$:Int (Combat Property), Right.Crew$:Int (Combat Property)
           Crew.
           This returns the raw, unfiltered value of the %Crew field within the VCR data structure.
           This field normally has a meaning only for ships.
           @see Crew (Combat Participant Property) */
        return makeIntValue(o.getCrew());
     case ivsDamage:
        /* Damage documented in getShipProperty */
        /* @q Left.Damage:Int (Combat Property), Right.Damage:Int (Combat Property)
           Initial damage in percent. */
        return makeIntValue(o.getDamage());
     case ivsId:
        /* @q Id:Int (Combat Participant Property)
           @q Left.Id:Int (Combat Property), Right.Id:Int (Combat Property)
           Id of this ship or planet. */
        return makeIntValue(o.getId());
     case ivsMass:
        /* @q Mass:Int (Combat Participant Property)
           @q Left.Mass:Int (Combat Property), Right.Mass:Int (Combat Property)
           Combat mass of this unit.
           This mass includes the hull weight and optional bonuses, such as the Engine-Shield-Bonus,
           but not the ship's cargo, equipment or ammo. It therefore cannot be meaningfully compared
           to a {Mass (Ship Property)|ship's mass}. */
        return makeIntValue(o.getMass());
     case ivsName:
        /* @q Name:Int (Combat Participant Property)
           @q Left.Name:Int (Combat Property), Right.Name:Int (Combat Property)
           Name of ship or planet. */
        return makeStringValue(o.getName());
     case ivsNameFull:
        /* @q Name.Full:Int (Combat Participant Property)
           @q Left:Int (Combat Property), Right:Int (Combat Property)
           Name and type of this combat participant.
           A string of the form "name (Planet #Id)" resp. "name (Ship #Id)". */
        return makeStringValue(format(o.isPlanet()
                                      ? _("%s (Planet #%d)")
                                      : _("%s (Ship #%d)"),
                                      o.getName(),
                                      o.getId()));
     case ivsOwnerAdj:
        /* @q Owner.Adj:Str (Combat Participant Property)
           @q Left.Owner.Adj:Str (Combat Property), Right.Owner.Adj:Str (Combat Property)
           Adjective name of this player. */
        if (o.getOwner() > 0 && o.getOwner() <= NUM_OWNERS) {
            return makeStringValue(player_racenames.getAdjName(o.getOwner()));
        } else {
            return 0;
        }
     case ivsOwnerId:
        /* @q Owner$:Int (Combat Participant Property)
           @q Left.Owner$:Int (Combat Property), Right.Owner$:Int (Combat Property)
           Player number. */
        return makeIntValue(o.getOwner());
     case ivsOwnerShort:
        /* @q Owner:Str (Combat Participant Property)
           @q Left.Owner:Str (Combat Property), Right.Owner:Str (Combat Property)
           Short name of this player. */
        if (o.getOwner() > 0 && o.getOwner() <= NUM_OWNERS) {
            return makeStringValue(player_racenames.getShortName(o.getOwner()));
        } else {
            return 0;
        }
     case ivsShield:
        /* @q Shield:Int (Combat Participant Property)
           @q Left.Shield:Int (Combat Property), Right.Shield:Int (Combat Property)
           Initial shield level in percent. */
        return makeIntValue(o.getShield());
     case ivsType:
     case ivsTypeShort:
        /* @q Type:Str (Combat Participant Property)
           @q Left.Type:Str (Combat Property), Right.Type:Str (Combat Property)
           Classification of this unit. Possible values are:
           - "Planet"
           - "Carrier"
           - "Torpedo Ship"
           - "Beam Weapons"
           - "Freighter" */
        /* @q Type.Short:Str (Combat Participant Property)
           @q Left.Type.Short:Str (Combat Property), Right.Type.Short:Str (Combat Property)
           Classification of this unit, short.
           This is the first letter of the {Type (Combat Participant Property)|Type}, see there. */
        {
            const char* kind;
            if (o.isPlanet()) {
                kind = "Planet";
            } else if (o.getBayCount() > 0) {
                kind = "Carrier";
            } else if (o.getTorpLauncherCount() > 0 && o.getTorpType() > 0 && o.getTorpType() <= NUM_TORPS) {
                kind = "Torpedo Ship";
            } else if (o.getBeamCount() > 0) {
                kind = "Beam Weapons";
            } else {
                kind = "Freighter";
            }
            if (ivs == ivsTypeShort) {
                return makeStringValue(string_t(kind, 1));
            } else {
                return makeStringValue(kind);
            }
        }
        break;
     case ivsHullName:
     case ivsHullId:
        /* @q Hull:Str (Combat Participant Property)
           @q Left.Hull:Str (Combat Property), Right.Hull:Str (Combat Property)
           Hull name.
           EMPTY if the hull cannot be determined, or this is a planet. */
        /* @q Hull$:Int (Combat Participant Property)
           @q Left.Hull$:Int (Combat Property), Right.Hull$:Int (Combat Property)
           Hull number.
           EMPTY if the hull cannot be determined, or this is a planet. */
        if (int h = o.getGuessedHull()) {
            if (ivs == ivsHullId) {
                return makeIntValue(h);
            } else {
                return makeStringValue(::getHull(h).getName());
            }
        } else {
            return 0;
        }
     case ivsImage:
        /* @q Image:Int (Combat Participant Property)
           @q Left.Image:Int (Combat Property), Right.Image:Int (Combat Property)
           Number of ship picture.
           If the hull of the ship can be determined, your changes to the picture assignment will be honored,
           otherwise, the host-provided picture is returned.
           For planets, this property is 0. */
        return makeIntValue(o.getGuessedShipPicture());
     case ivsLevel:
        /* @q Level:Int (Combat Participant Property)
           @q Left.Level:Int (Combat Property), Right.Level:Int (Combat Property)
           Experience level for this unit.
           0 if the fight does not include experience levels (because experience is not enabled, maybe). */
        return makeIntValue(o.getExperienceLevel());
     case ivsStatus:
        /* @q Status:Str (Combat Participant Property)
           @q Left.Status:Str (Combat Property), Right.Status:Str (Combat Property)
           Battle result, from the point-of-view of this unit.
           - "Survived" if the unit survived the battle and captured or destroyed its opponent.
           - "Captured" if the unit was captured by an enemy.
           - "Exploded" if the unit was destroyed.
           - "Invalid" if the battle is not playable in PCC.
           - "Stalemate" if the battle ended without a victor (PCC2 only).
           - "Disabled" if the battle ended because combatants didn't have offensive capabilities left (PCC 1.x only).
           - "Timeout" if the battle timed out (PCC 1.x only).

           Computing the value for this property may involve playing the whole VCR,
           and thus take a considerable amount of time.
           Results are cached, so you'll only have to wait once. */
        return makeVcrResult(vcr, side);
     case ivsStatusRaw:
        /* @q Status$:Int (Combat Participant Property)
           @q Left.Status$:Int (Combat Property), Right.Status$:Int (Combat Property)
           Battle result, from the point-of-view of this unit.
           This is an integer:
           - -1: this unit was destroyed.
           - 0: this unit survived the battle.
           - other: this unit was captured in battle, the value is the new owner's player number. */
        return makeIntValue(vcr.getOutcome(side));
     case ivsIsPlanet:
        /* @q IsPlanet:Bool (Combat Participant Property)
           True if this is a planet. */
        return makeBoolValue(o.isPlanet());

     case ivsBeamKillRate:
        /* @q Config.BeamKillRate:Int (Combat Participant Property)
           Beam kill rate for this unit (3 for Privateers, otherwise 1).
           @since PCC2 1.99.23 */
        return makeIntValue(o.getBeamKillRate());

     case ivsBeamChargeRate:
        /* @q Config.BeamChargeRate:Int (Combat Participant Property)
           Beam charge rate boost. This value is generated only by NuHost.
           In particular, it is not used by PHost.
           It can be used for PHost in PCC2's simulator, where it scales up the effective BeamRechargeRate
           computed from PConfig.
           @since PCC2 1.99.23 */
        return makeIntValue(o.getBeamChargeRate());

     case ivsTorpMissRate:
        /* @q Config.TorpMissRate:Int (Combat Participant Property)
           Torpedo miss rate. This value is generated only by NuHost.
           In particular, it is not used by PHost and has no relation to the TorpHitOdds PConfig option.
           @since PCC2 1.99.23 */
        return makeIntValue(o.getTorpMissRate());

     case ivsTorpChargeRate:
        /* @q Config.TorpChargeRate:Int (Combat Participant Property)
           Torpedo charge rate boost. This value is generated only by NuHost.
           In particular, it is not used by PHost.
           It can be used for PHost in PCC2's simulator, where it scales up the effective TubeRechargeRate
           computed from PConfig.
           @since PCC2 1.99.23 */
        return makeIntValue(o.getTorpChargeRate());

     case ivsCrewDefenseRate:
        /* @q Config.CrewDefenseRate:Int (Combat Participant Property)
           Crew defense rate. This value is generated only by NuHost.
           In particular, it is not used by PHost.
           It can be used for PHost in PCC2's simulator, where it scales down the effective CrewKillScaling
           computed from PConfig (a CrewDefenseRate of 100 reduces the CrewKillScaling to 0).
           @since PCC2 1.99.23 */
        return makeIntValue(o.getCrewDefenseRate());
    }
    return 0;
}

/** Get property of a VCR record.
    \param vcr The record
    \param ivp Property to get */
IntValue*
getVcrProperty(GVcrEntry& vcr, IntVcrProperty ivp)
{
    GClassicVcrEntry* ce = dynamic_cast<GClassicVcrEntry*>(&vcr);
    switch (ivp) {
     case ivpSeed:
        /* @q Seed:Int (Combat Property)
           Random number seed.
           Valid only for classic combat, EMPTY for others. */
        if (ce) {
            return makeIntValue(ce->getVcr().getSeed());
        } else {
            return 0;
        }
     case ivpMagic:
        /* @q Magic:Int (Combat Property)
           VCR algorithm identification value.
           Valid only for classic combat, EMPTY for others. */
        if (ce) {
            return makeIntValue(ce->getVcr().getSignature());
        } else {
            return 0;
        }
     case ivpType:
        /* @q Type$:Int (Combat Property)
           Unit type identification value.
           - 0: this is a ship/ship fight.
           - 1: this is a ship/planet fight, {Right (Combat Property)|Right} resp.
                {Unit (Combat Property)|Unit(1)} is a planet.
           Valid only for classic combat, EMPTY for others. */
        if (ce) {
            return makeIntValue(ce->getVcr()(1).isPlanet());
        } else {
            return 0;
        }
     case ivpAlgorithm:
        /* @q Algorithm:Str (Combat Property)
           Name of VCR algorithm. */
        return makeStringValue(vcr.getAlgorithmName(false));
     case ivpFlags:
        /* @q Capabilities:Int (Combat Property)
           VCR feature identification value.
           Valid only for classic combat, EMPTY for others. */
        if (ce) {
            return makeIntValue(ce->getCapabilities());
        } else {
            return 0;
        }
     case ivpNumUnits:
        /* @q NumUnits:Int (Combat Property)
           Number of units participating in this fight.
           This is the number of elements in the {Unit (Combat Property)|Unit} array.
           @since PCC2 1.99.19 */
        return makeIntValue(vcr.getNumObjects());
     case ivpUnits:
        /* @q Unit:Obj() (Combat Property)
           Information about all participating units.
           Each object in this array has {int:index:group:combatparticipantproperty|Combat Participant Properties}.
           Indexes are 1 to {NumUnits (Combat Property)|NumUnits}.

           The properties of <tt>Units(1)</tt> and <tt>Units(2)</tt> are also available as Combat Properties
           %Left.XXX and %Right.XXX, mainly for classic 1:1 combat.

           @since PCC2 1.99.19 */
        return new IntVcrSideArray(vcr);
    }
    return 0;
}
