/**
  *  \file testmsgt.cc
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/msgparse.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/console.h"
#include "util/modinit.h"
#include "arch/platform.h"
#include "game/racename.h"
#include "game/gamedirs.h"

static const ModuleInitializer* testmsgt_modules[] = {
    mod_base_console,
    mod_racenames,
    0,
};

int main(int argc, char** argv)
{
    performInitialisation(OnStartup, testmsgt_modules);
    if (argc < 3) {
        console << "testmsgt mode msgparse.ini [gamedir]\n";
        return 1;
    }

    if (argc >= 4) {
        initDirectories(argv[3], Platform::getDataDirectory());
        performInitialisation(OnEnterDirectory, testmsgt_modules);
    }

    try {
        Ptr<Stream> s = Stream::openFile(argv[2], Stream::C_READ);
        GMessageParser mp;
        mp.load(*s);
        if (strcmp(argv[1], "dump") == 0) {
            mp.dump();
        } else if (strcmp(argv[1], "raw") == 0) {
            // Build a message
            GMessageLines lines;
            string_t line;
            while (Platform::readConsoleLine("", line))
                lines.push_back(line);
            
            // Parse it
            for (uint32_t i = 0; i < mp.getNumTemplates(); ++i) {
                std::vector<string_t> result;
                if (mp.getTemplate(i).match(lines, 12, result)) {
                    console << "This message matches the following template:\n------------\n";
                    mp.getTemplate(i).dump();
                    console << "-----------\nwith values:\n";
                    for (uint32_t j = 0; j < result.size(); ++j)
                        console << "  " << j << " '" << mp.getTemplate(i).getVariableName(j) << "' -> '" << result[j] << "'\n";
                    if (int(result.size()) != mp.getTemplate(i).getNumVariables())
                        console << "Variable count mismatch\n";
                }
            }
        } else if (strcmp(argv[1], "cooked") == 0) {
            // Build a message
            string_t msg, line;
            while (Platform::readConsoleLine("", line))
                msg += line, msg += '\n';

            ptr_vector<GMessageInformation> mi;
            mp.parseMessage(msg, 12, 42, mi);
            if (mi.empty()) {
                console << "Nothing found.\n";
            } else {
                for (uint_t i = 0; i < mi.size(); ++i) {
                    const GMessageInformation& m = *mi[i];
                    console << "Object type " << m.getObjectType() << " #" << m.getObjectId() << ", turn " << m.getTurnNumber() << ":\n";
                    for (GMessageInformation::iterator j = m.begin(); j != m.end(); ++j) {
                        if (GMessageConfigValue* cv = dynamic_cast<GMessageConfigValue*>(*j)) {
                            console << "  config " << cv->getIndex() << " = " << cv->getValue() << "\n";
                        } else if (GMessageIntValue* iv = dynamic_cast<GMessageIntValue*>(*j)) {
                            console << "  " << getNameForItem(iv->getIndex()) << " = " << iv->getValue() << "\n";
                        } else if (GMessageStringValue* sv = dynamic_cast<GMessageStringValue*>(*j)) {
                            console << "  " << getNameForItem(sv->getIndex()) << " = " << sv->getValue() << "\n";
                        } else if (GMessageScoreValue* scv = dynamic_cast<GMessageScoreValue*>(*j)) {
                            console << "  score " << scv->getIndex() << " = " << scv->getValue() << "\n";
                        } else if (GMessageAllianceValue* av = dynamic_cast<GMessageAllianceValue*>(*j)) {
                            static const char OFFERS[] = "N.CY";
                            console << "  alliance " << av->getIndex() << "\n";
                            const GAllianceOffer& offer = av->getValue();
                            console << "     their offers: ";
                            for (int i = 1; i <= 11; ++i) console << OFFERS[offer.theirOffer[i]];
                            console << "\n";
                            console << "     old offers:   ";
                            for (int i = 1; i <= 11; ++i) console << OFFERS[offer.oldOffer[i]];
                            console << "\n";
                            console << "     new offers:   ";
                            for (int i = 1; i <= 11; ++i) console << OFFERS[offer.newOffer[i]];
                            console << "\n";
                        } else {
                            console << "  unknown value\n";
                        }
                    }
                }
            }
        } else {
            console << "mode must be dump/raw/cooked\n";
        }
    }
    catch (FileProblemException& e) {
        console << e.getFileName() << ": " << e.what() << "\n";
    }
    catch (...) {
        console << "Exception!\n";
    }
    return 0;
}
