/**
  *  \file ui/msgbox.cc
  *  \brief Implementation of Standard Message Boxes
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/msgbox.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/richtext.h"
#include "ui/widgets/richstatic.h"
#include "ui/widgets/quit.h"

/******************************* WDialogBox ******************************/

UIMessageBox::UIMessageBox(const char* text, string_t title)
    : UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY, 10),
      group(h.add(new UIGroup(UIHBoxLayout::instance5))),
      trans(h.add(new UIKeyTranslator())),
      last_command(cm_Escape)
{
    init(RichText(text));
}

UIMessageBox::UIMessageBox(string_t text, string_t title)
    : UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY, 10),
      group(h.add(new UIGroup(UIHBoxLayout::instance5))),
      trans(h.add(new UIKeyTranslator())),
      last_command(cm_Escape)
{
    init(RichText(text));
}

UIMessageBox::UIMessageBox(const RichText& text, string_t title)
    : UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY, 10),
      group(h.add(new UIGroup(UIHBoxLayout::instance5))),
      trans(h.add(new UIKeyTranslator())),
      last_command(cm_Escape)
{
    init(text);
}

void
UIMessageBox::init(const RichText& text)
{
    setLayout(h.add(new UIVBoxLayout(10)));

    RichText bigText(text);
    bigText.withAttribute(new RichTextStyleAttribute(RichTextStyleAttribute::Big));

    RichDocument tmp;
    tmp.setPageWidth(440);
    tmp.add(bigText);
    tmp.finish();
    int width = tmp.getDocumentWidth();
    if (width < 200) {
        width = 200;
    }
    add(h.add(new UIRichStatic(bigText, width)));
    add(group);
    add(trans);

    group.add(h.add(new UISpacer()));

    addKey(cm_Escape, SDLK_ESCAPE);
    addKey(cm_Enter,  SDLK_RETURN);
}

UIMessageBox&
UIMessageBox::addButton(int id, string_t btn, UIKey k)
{
    group.add(h.add(new UIButton(id, btn, k)).addStop());
    last_command = id;
    return *this;
}

UIMessageBox&
UIMessageBox::addKey(int id, UIKey k)
{
    trans.addStop(k, id);
    return *this;
}

int
UIMessageBox::run()
{
    group.add(h.add(new UISpacer()));
    group.add(h.add(new UIQuit(last_command)));
    pack();

    ui_root->centerWidget(*this);
    return ui_root->runModal(*this);
}

bool
UIMessageBox::doYesNoDialog()
{
    UIKeyString y(_("Yes")), n(_("No"));
    addButton(cm_Enter,  y.getString(), y.getKey());
    addButton(cm_Escape, n.getString(), n.getKey());
    addKey(cm_Enter, ' ');
    return run() == cm_Enter;
}

void
UIMessageBox::doOkDialog()
{
    addButton(cm_Enter, _("OK"), ' ');
    run();
}

/** Standard Message Box. A message box containing the specified text
    and title, and an "OK" button. The message box is modal. The text
    can contain '\n' to start new lines; it is automatically
    word-wrapped. */
void
messageBox(string_t text, string_t title)
{
    return UIMessageBox(text, title).doOkDialog();
}
